#import "PipeTest.h"

#import <process.h>
#import <windows.h>
#include <shellapi.h>
#include <stdio.h>

#import <Foundation/NSData.h>
#import <Foundation/NSProcessInfo.h>
#import <Foundation/NSNotification.h>
#import <Foundation/NSPathUtilities.h>
#import <Foundation/NSFileHandle.h>
#import <Foundation/NSRunLoop.h>

@implementation PipeTest

- init
{
    STARTUPINFO startup;
    PROCESS_INFORMATION pinfo;
    SECURITY_ATTRIBUTES sa;
    HANDLE inFile=0;
    HANDLE hOutPipeRead = 0, hOutPipeWrite = 0;
    HANDLE hErrPipeRead = 0, hErrPipeWrite = 0;

    [ super init ];

    sa.nLength = sizeof(sa);
    sa.bInheritHandle = TRUE;
    sa.lpSecurityDescriptor = NULL;

    inFile = CreateFile("NUL:", GENERIC_READ, FILE_SHARE_READ, &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (inFile == INVALID_HANDLE_VALUE) {
        return 0;
    }


    CreatePipe(&hOutPipeRead, &hOutPipeWrite, &sa, 0);
    stdOut = [[ NSFileHandle alloc] initWithNativeHandle: hOutPipeRead
                closeOnDealloc: YES];
    CreatePipe(&hErrPipeRead, &hErrPipeWrite, &sa, 0);
    stdErr = [[ NSFileHandle alloc] initWithNativeHandle: hErrPipeRead
                closeOnDealloc: YES];

    [[ NSNotificationCenter defaultCenter ]
        addObserver: self
           selector: @selector(dataNeedsPosting: )
               name: NSFileHandleReadCompletionNotification
             object: stdOut ];

    [[ NSNotificationCenter defaultCenter ]
        addObserver: self
           selector: @selector(dataNeedsPostingErr: )
               name: NSFileHandleReadCompletionNotification
             object: stdErr ];

    startup.cb=sizeof(startup);
    startup.lpReserved = NULL;
    startup.lpDesktop = NULL;
    startup.lpTitle = NULL;
    startup.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
    startup.wShowWindow = SW_SHOWNORMAL;
    startup.cbReserved2 = 0;
    startup.lpReserved2 = NULL;
    startup.hStdInput = inFile;
    startup.hStdOutput = hOutPipeWrite;
    startup.hStdError = hErrPipeWrite;

    if (!CreateProcess(0, "yammer", NULL, NULL, TRUE, CREATE_DEFAULT_ERROR_MODE | NORMAL_PRIORITY_CLASS, 0, 0, &startup, &pinfo)) {
        char * lpvMessage;
        FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
                            FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(),
                            MAKELANGID (LANG_ENGLISH, SUBLANG_ENGLISH_US),
                            (LPTSTR)&lpvMessage, 0, NULL);

        printf("Task launch failed!\n%s", lpvMessage);
        [stdOut release];
        stdOut=0;
        [stdErr release];
        stdErr=0;
        CloseHandle(hOutPipeWrite);
        CloseHandle(hErrPipeWrite);
        CloseHandle(inFile);
        return 0;
    }
    processHandle = pinfo.hProcess;
    CloseHandle(pinfo.hThread);

    [[NSRunLoop currentRunLoop] addEvent:processHandle type:ET_HANDLE watcher:self forMode:0];

    [ stdOut readInBackgroundAndNotify ];
    [ stdErr readInBackgroundAndNotify ];

    CloseHandle(hOutPipeWrite);
    CloseHandle(hErrPipeWrite);
    CloseHandle(inFile);
    return self;
}

- (void) dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver: self];
    if (processHandle)
        TerminateProcess( processHandle , 1 );
    [ stdOut release ];
    stdOut = 0;
    [ stdErr release ];
    stdErr = 0;
    [ super dealloc ];
}

- (void) receivedEvent: (void*)data type: (RunLoopEventType)type extra: (void*)extra forMode: (NSString*)mode
{
    if (!processHandle)
        return;
    if (WAIT_TIMEOUT != WaitForSingleObject(processHandle, 0))
        [self taskTerminatedNotification: 0];
    [[NSRunLoop currentRunLoop] removeEvent:data type:ET_HANDLE forMode:mode all:YES];
}


- (void) taskTerminatedNotification: (NSNotification*)note
{
    DWORD status = 0;
    if (!GetExitCodeProcess(processHandle, &status))
        return;
    if (status != STILL_ACTIVE) {
        [ self postData: [ stdOut readDataToEndOfFile ]];
        [ self postData: [ stdErr readDataToEndOfFile ]];
    }
    [ stdOut closeFile ];
    [ stdOut release ];
    stdOut = 0;
    [ stdErr closeFile ];
    [ stdErr release ];
    stdErr = 0;
    CloseHandle(processHandle);
    processHandle = 0;
}


- (void) dataNeedsPosting: (NSNotification*)note
{
    NSData *data = [[ note userInfo ] objectForKey:NSFileHandleNotificationDataItem ];
    if (data && [data length]) {
        [ self postData: data ];
        [ stdOut readInBackgroundAndNotify ];
    }
}

- (void) dataNeedsPostingErr: (NSNotification*)note
{
    NSData *data = [[ note userInfo ] objectForKey:NSFileHandleNotificationDataItem ];
    if (data && [data length]) {
        [ self postData: data ];
        [ stdErr readInBackgroundAndNotify ];
    }
}

- (void) postData: (NSData *) data
{
    if (data && [data length]) {
        fwrite([data bytes], 1, [data length], stdout);
    }
}

@end

